# $Id: mp3info.rb,v 1.2 2003/03/18 19:21:55 moumar Exp $
# = Description
#
# ruby-mp3info gives you access to low level informations on mp3 files
# (bitrate, length, samplerate, etc...). It is written in pure ruby.
# This library doesn't fetch ID3 tags. For that, check raa for mp3tag
#
#
# = Download
#
# get tar.gz at
# http://davedd.free.fr/ruby-mp3info/download/
#
#
# = Installation
# 
#   $ ruby install.rb config
#   $ ruby install.rb setup
#   # ruby install.rb install
#
#
# = Example
# 
#   require "mp3info"
#
#   mp3info = Mp3Info.new("myfile.mp3")
#   puts "MPEG #{info.mpeg_version} Layer #{info.layer} #{info.vbr ? "VBR" : "CBR"} #{info.bitrate} Kbps
#      #{info.channel_mode} #{info.samplerate} Hz length #{info.length} sec.
#      error protection #{info.error_protection}"
#
#
# = Testing
#   
# Test::Unit library is used for tests. see http://testunit.talbott.ws/
#
#  $ ruby test.rb
#
#
# = ToDo
#
# * More test (soon)
# * adding support for ID3v* tags (that might improve speed)
#
# = Changelog
#
# [0.1 17/03/2003]
#
# * Initial version
#
#
# License:: Ruby
# Author:: Guillaume Pierronnet (mailto:moumar@netcourrier.com)
# Website:: http://davedd.free.fr/ruby-mp3info/

# Raised on any kind of error related to ruby-mp3info
class Mp3InfoError < StandardError ; end

class Mp3InfoInternalError < StandardError ; end #:nodoc:

class Mp3Info

  VERSION = "0.1"

  MPEG_VERSION = [ 2, nil, 2, 1]
  LAYER = [ nil, 3, 2, 1]
  BITRATE = [ 
    [ 
      [32, 64, 96, 128, 160, 192, 224, 256, 288, 320, 352, 384, 416, 448],
      [32, 48, 56, 64, 80, 96, 112, 128, 160, 192, 224, 256, 320, 384],
      [32, 40, 48, 56, 64, 80, 96, 112, 128, 160, 192, 224, 256, 320] ],
    [ 
      [32, 48, 56, 64, 80, 96, 112, 128, 144, 160, 176, 192, 224, 256],
      [8, 16, 24, 32, 40, 48, 56, 64, 80, 96, 112, 128, 144, 160],
      [8, 16, 24, 32, 40, 48, 56, 64, 80, 96, 112, 128, 144, 160]
    ]
  ]
  SAMPLERATE = [
    [ 44100, 48000, 32000 ],
    [ 22050, 24000, 16000 ]
  ]
  CHANNEL_MODE = [ "Stereo", "JStereo", "Dual Channel", "Single Channel"]

  # mpeg version = 1 or 2
  attr_reader(:mpeg_version)

  # layer = 1, 2, or 3
  attr_reader(:layer)
  
  # bitrate in kbps
  attr_reader(:bitrate)

  # samplerate in Hz
  attr_reader(:samplerate)
  
  # channel mode = "Stereo", "JStereo", "Dual Channel" or "Single Channel"
  attr_reader(:channel_mode)

  # variable bitrate = true or false
  attr_reader(:vbr)

  # length in seconds
  attr_reader(:length)

  # error protection = true or false
  attr_reader(:error_protection)

  def initialize(filename)
    raise Mp3InfoError, "empty file" unless File.stat(filename).size? #FIXME
    @vbr = false
    @file = File.open(filename)
    begin
      a = find_frame_sync
      b, c = @file.read(4).unpack("C2")
      mpeg_version = (a >> 3) & 3
      @mpeg_version = MPEG_VERSION[mpeg_version]
      raise Mp3InfoInternalError if @mpeg_version.nil?
      layer = (a >> 1) & 3
      @layer = LAYER[layer]
      raise Mp3InfoInternalError if @layer.nil?
      @error_protection = a & 1 == 1 ? true : false
      bitrate = b >> 4
      @bitrate = BITRATE[@mpeg_version-1][@layer-1][bitrate-1] or raise Mp3InfoInternalError
      samplerate = (b >> 2) & 3
      @samplerate = SAMPLERATE[@mpeg_version-1][samplerate] or raise Mp3InfoInternalError
  #    @padding_bit = (b >> 1) & 1
      channel_mode = c >> 6
      @channel_mode = CHANNEL_MODE[channel_mode]

      seek = 
      if mpeg_version == 3			# mpeg version 1
        if channel_mode == 3 
	  15 					# Single Channel
        else
	  30
	end
      else					# mpeg version 2 or 2.5
        if channel_mode == 3 			# Single Channel
	  7	
        else
	  15
	end
      end

      @file.seek(seek, File::SEEK_CUR)

      if @file.read(4) == "Xing"			# This is a VBR file
        @vbr = true
        @file.seek(4, File::SEEK_CUR)			#we have the frames number after
	#if @file.read(4).unpack("N")[0] == 0xF
	#end
	@frames = @file.read(4).unpack("N")[0]
	medframe = @file.stat.size / @frames.to_f
	@bitrate = ( (medframe * @samplerate) /  ( 1000 * ( layer==3 ? 12 : 144) ) ).to_i
      end

      @length = (( (8 * @file.stat.size) / 1000) / @bitrate).to_i

    rescue Mp3InfoInternalError
      retry
    ensure
      @file.close
    end
  end
  
private
  def find_frame_sync
    last_byte = 0
    while b = @file.read(1)
      byte = b.unpack("C")[0]
    #@file.each_byte { |b| 
      return byte if byte >> 5 == 7 and last_byte == 0xFF
      raise Mp3InfoError if @file.pos > 200000 or @file.eof?
      last_byte = byte
    #}
    end
  end
 
end

if $0 == __FILE__
  while filename = ARGV.shift
    begin
      info = Mp3Info.new(filename)
      puts filename
      puts "MPEG #{info.mpeg_version} Layer #{info.layer} #{info.vbr ? "VBR" : "CBR"} #{info.bitrate} Kbps \
      #{info.channel_mode} #{info.samplerate} Hz length #{info.length} sec."
    rescue Mp3InfoError => e
      puts "#{filename}\nERROR: #{e}"
    end
    puts 
  end
end
